# The reason for the strange code is that I want the function to
# respect the dimensions of theta, ie if theta is a matrix then return
# a matrix. 

# The reason for hard coding these rather than relying on extracting
# them from the familytoo object is that it simply seems faster this
# way.



gauscumone   <-  function(theta,...){theta}
gauscumtwo   <-  function(theta,...){theta * 1/theta * 1}
gauscumthree <-  function(theta,...){theta * 0}
gauscumfour  <-  function(theta,...){theta * 0}

bincumone   <-  function(theta,n=1){n * theta}
bincumtwo   <-  function(theta,n=1){n * theta * (1 - theta)}
bincumthree <-  function(theta,n=1){n * theta * (1 - theta) * ( 1 - 2 * theta)}
bincumfour  <-  function(theta,n=1){n * theta * (1 - theta) * ( 1 - 6 * theta * (1 - theta))}

poiscumone   <-  function(theta,...){theta}
poiscumtwo   <-  function(theta,...){theta}
poiscumthree <-  function(theta,...){theta}
poiscumfour  <-  function(theta,...){theta}

gamcumone    <- function(theta,...){theta}
gamcumtwo    <- function(theta,...){theta^2}
gamcumthree  <- function(theta,...){2 * theta^3}
gamcumfour   <- function(theta,...){6 * theta^4}



# A function to check for convergence defined as a small relative
# change in parameter values.
convergecheck <- function(a){
top <- abs(a[1] - a[2])
bot <- abs(a[1])
if(abs(bot) > 1e-8) return(top/bot) else return(0)
}

# Return the sum of the elements of a hadamard product of two matrices.

hadsum <- function(a,b){return(sum(a*b))}

# Cumulant calculation function.
cumcalc <- function(x,psi,lambda,fam,nu){

# use familytoo to make the mean function
meanfunc <- familytoo(fam)$inverse

mu <- (if(familytoo(fam)$family[1] != "Gamma") 1 else -1) * 
meanfunc(matrix(lambda,ncol=1) %*% matrix(rep(1,dim(x)[2]),nrow=1)  + 
if(dim(x)[2] < dim(x)[1]) {apply(x,c(2),"%*%",psi)}
else{t(apply(x,c(1),"%*%",psi))}
)


## Here is what I planned on doing. But it was much slower than
## the if statments below.

#cone <-   familytoo(fam)$First
#ctwo <-   familytoo(fam)$Second
#cthree <- familytoo(fam)$Third
#cfour <-  familytoo(fam)$Fourth 
#cumone   <- cone(mu,nu)
#cumtwo   <- ctwo(mu,nu)
#cumthree <- cthree(mu,nu)
#cumfour  <- cfour(mu,nu)

if(family(fam)$family[1] == "Gaussian"){
cumone   <- gauscumone(mu,nu) 
cumtwo   <- gauscumtwo(mu,nu) 
cumthree <- gauscumthree(mu,nu)
cumfour  <- gauscumfour(mu,nu) 
}


if(family(fam)$family[1] == "Binomial"){
cumone   <- bincumone(mu,nu) 
cumtwo   <- bincumtwo(mu,nu) 
cumthree <- bincumthree(mu,nu)
cumfour  <- bincumfour(mu,nu) 
}

if(family(fam)$family[1] == "Poisson"){
cumone   <- poiscumone(mu)
cumtwo   <- poiscumtwo(mu)
cumthree <- poiscumthree(mu)
cumfour  <- poiscumfour(mu)
}

if(family(fam)$family[1] == "Gamma"){
cumone   <- gamcumone(mu,nu)
cumtwo   <- gamcumtwo(mu,nu)
cumthree <- gamcumthree(mu,nu)
cumfour  <- gamcumfour(mu,nu)
}

a <- list(cumone,cumtwo,cumthree,cumfour)
return(a)
}

# The correction function.
correctiona <- function(w,cumtwo,cumthree,cumfour){
	rowcumtwo   <- c(( cumtwo) %*% rep.int(1,ncol(cumtwo)))
	rowcumthree <- c(( cumthree) %*% rep.int(1,ncol(cumthree)))
	rowcumfour  <- c(( cumfour) %*% rep.int(1,ncol(cumfour)))
	a <- 1/(rowcumtwo * rowcumfour + 2 * rowcumtwo^3 -
		rowcumthree^2) * cbind(- rowcumtwo * rowcumthree,
		rowcumtwo^2)
	return(a)
}


# This is the U2 score function -- the projected score.
utwoa <- function(z,y,w,cumone,cumtwo,cumthree,cumfour){
	corr <- correctiona(w,cumtwo,cumthree,cumfour)
	a <- apply(z,c(3),hadsum,((y - cumone) + cumtwo * corr[,1] +
		cumthree * corr[,2]) *w) 
	return(a)
}

# This is the U1 score function -- the mle score.
uonea <- function(z,y,w,cumone){
	a <- apply(z,c(3),hadsum,((y - cumone) *w)) 
	return(a)
}


# The expected info matrix. When I have more time I'll do this more
# elegantly.

expinfoa <- function(z,w,cumtwo){
        params <- dim(z)[3]
	info <- matrix(rep(0,params^2),params)
	for (i in 1:params){
		for (j in 1:params){
			info[[i,j]] <- sum(z[,,i] * (cumtwo *w ) * z[,,j])
			}
		}
	return(info)
}

# Now a function that takes info, score and parameters and updates

newpsival <- function(info,score,param){
	return(solve(info, info %*% param + score))
}

# Do a similar one for the nuisance parameters
# Now the arguments are all vectors beacuse of the independent
# strata assumption.

newlambdaval <- function(info,score,param){
	return(param + score/info)
}



##############################################################
## Here are the numerical derivative calculations, which are #
## needed in the calculation of the efficient information.   #
##############################################################


# firstly D[U2,psi]

# The derivative with respect to psi of the U2 score function.

dutwopsi <- function(y,x,fam,w,psi,lambda,delta,nu){  
	dimpsi <- dim(x)[3]
	a <- matrix(rep(0,(dimpsi)^2),ncol = dimpsi)
	for( i in 1:dimpsi){
		psi[i] <- psi[i] + delta
		cumlist <- cumcalc(x,psi,lambda,fam,nu)
		a[,i] <- utwoa(x,y,w,cumlist[[1]],cumlist[[2]],
			cumlist[[3]],cumlist[[4]])
		psi[i] <- psi[i] - 2 * delta
		cumlist <- cumcalc(x,psi,lambda,fam,nu)
		a[,i] <- (a[,i] -   utwoa(x,y,w,cumlist[[1]],
			cumlist[[2]],cumlist[[3]],cumlist[[4]]))/(2 * delta)
		psi[i] <- psi[i] + delta
		}
# Only symmetrize this if dimension is greater than 1. 
	if(dimpsi > 1){
		for(i in 2:dimpsi){for(j in 1:i){
			a[i,j] <- a[j,i] <- (a[i,j] + a[j,i])/2
			}
		}
	}
return(a)
}


# Now D[U2,lambda]
# The derivative with respect to lambda of the U2 score

dutwolambda <- function(y,x,fam,w,psi,lambda,delta,nu){
	ro <- dim(x)[1]
	co <- dim(x)[2]
	po <- dim(x)[3]
	lambda <- lambda + delta
	cumlist <- cumcalc(x,psi,lambda,fam,nu)
	corr <- correctiona(w,cumlist[[2]],cumlist[[3]],cumlist[[4]])
	upper <- (( y - cumlist[[1]]) + cumlist[[2]] * corr[,1] +
	cumlist[[3]] * corr[,2]) * w
	upper <- apply(array(apply(x,c(3),"*",upper),c(ro,co,po)),c(1,3),sum)
	lambda <-  lambda - 2 * delta
	cumlist <- cumcalc(x,psi,lambda,fam,nu)
	corr <- correctiona(w,cumlist[[2]],cumlist[[3]],cumlist[[4]])
	lower <- (( y - cumlist[[1]]) + cumlist[[2]] * corr[,1] +
	cumlist[[3]] * corr[,2]) *w
	lower <- apply(array(apply(x,c(3),"*",lower),c(ro,co,po)),c(1,3),sum)
	lambda <- lambda + delta
	return(t((upper - lower)/(2 * delta)))
}

# The derivative of V1 with respect to psi.

dvonepsi <- function(y,x,fam,w,psi,lambda,delta,nu){
	ro <- dim(x)[1]
	co <- dim(x)[2]
	po <- dim(x)[3]
	a <- matrix(rep(0,(ro * po)),ncol = po)
	for( i in 1:po){
		psi[i] <- psi[i] + delta
		cumlist <- cumcalc(x,psi,lambda,fam,nu)
		a[,i] <-  matrix(apply((y - cumlist[[1]])*w,1,sum),ncol=1)
		psi[i] <- psi[i] - 2 * delta
		cumlist <- cumcalc(x,psi,lambda,fam,nu)
		a[,i] <- (a[,i] - matrix(apply((y -
		cumlist[[1]])*w,1,sum),ncol=1))/(2 * delta)
		psi[i] <- psi[i] + delta
	}
return(a)
}

# Put the derivatives together to calculate the efficient information.

effinfo <- function(y,x,fam,w,psi,lambda,delta,nu){

# Symmetrize
	temp <- (dutwolambda(y,x,fam,w,psi,lambda,delta,nu) +
      		t(dvonepsi(y,x,fam,w,psi,lambda,delta,nu)))/2

	cumlist <- cumcalc(x,psi,lambda,fam,nu)      
	return(dutwopsi(y,x,fam,w,psi,lambda,delta,nu)  +
		temp  %*% (t(temp) * 1/apply(cumlist[[2]] * w,1,sum)))
	}

# Efficient info function for U1 score.
effinfouone <- function(x,w,cumtwo){

	ro <- dim(x)[1]
	co <- dim(x)[2]
	po <- dim(x)[3]

	temp <- t(apply(array(apply(x,c(3),"*",cumtwo*w),c(ro,co,po)),
	c(1,3),sum))
	return(-expinfoa(x,w,cumtwo)  +
	temp  %*% (t(temp) * 1/apply(cumtwo * w,1,sum)))
}

